"""django_workflow_engine package interface.

The django_workflow_engine package user utilises these artefacts to build their own
custom workflow e.g:

# I want to build my workflow
from django_workflow_engine import Step, Workflow

# I want to catch workflow errors
from django_workflow_engine import WorkflowError, WorkflowNotAuthError

# I want to derive from base Task
from django_workflow_engine.tasks import Task

# I want to use some built-in tasks
from django_workflow_engine.tasks import SendEmail, EmailFormTask, TaskError
"""
# Workflow definition
from .dataclass import Step
from .dataclass import Workflow

# Workflow execution
from .exceptions import WorkflowError
from .exceptions import WorkflowNotAuthError

# Framework built-ins
from .tasks import SendEmail
from .tasks import EmailFormTask
from .tasks import Task, TaskError



import uuid

from django.conf import settings
from django.shortcuts import reverse
from django.contrib.auth import get_user_model
from django.db import models

from django_workflow_engine.exceptions import WorkflowImproperlyConfigured

if not settings.DJANGO_WORKFLOWS:
    raise WorkflowImproperlyConfigured("Add DJANGO_WORKFLOWS to your settings")


class Flow(models.Model):
    WORKFLOWS = [
        (flow, flow.title().replace("_", " "))
        for flow in settings.DJANGO_WORKFLOWS
    ]

    workflow_name = models.CharField(
        "Select workflow", choices=WORKFLOWS, max_length=255
    )
    flow_name = models.CharField(
        "Select activity",
        max_length=255,
        help_text="e.g. Hire a Python Developer",
    )
    executed_by = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
    )
    started = models.DateTimeField(null=True)
    finished = models.DateTimeField(null=True, blank=True)
    flow_info = models.JSONField(default=dict)

    @property
    def is_complete(self):
        return bool(self.finished)

    @property
    def workflow(self):
        return settings.DJANGO_WORKFLOWS[self.workflow_name]

    @property
    def current_task_record(self):
        return self.tasks.filter(finished_at__isnull=True).first()

    # TODO: rename
    @property
    def nice_name(self):
        if self.is_complete:
            return "Completed"
        elif self.started:
            return "In-progress"
        else:
            return "Not started"

    @property
    def on_manual_step(self):
        if not self.current_task_record:
            return False

        current_step = self.workflow.get_step(self.current_task_record.step_id)

        if not current_step:
            return False

        return not self.workflow.get_step(self.current_task_record.step_id).task.auto

    @property
    def continue_url(self):
        return reverse("flow-continue", args=[self.pk])


class TaskRecord(models.Model):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    started_at = models.DateTimeField(auto_now_add=True)
    finished_at = models.DateTimeField(null=True)
    executed_by = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    flow = models.ForeignKey(Flow, on_delete=models.CASCADE, related_name="tasks")
    step_id = models.CharField(max_length=100)
    task_name = models.CharField(max_length=100)
    target = models.CharField(
        max_length=100,
        blank=True,
        null=True,
    )
    task_info = models.JSONField(default=dict)

    def __str__(self):
        return f"{self.step_id} {self.task_name}"


class TaskLog(models.Model):
    logged_at = models.DateTimeField(auto_now_add=True)
    message = models.CharField(max_length=255)
    task_record = models.ForeignKey(
        TaskRecord,
        related_name="log",
        on_delete=models.CASCADE,
    )



"""django_workflow_engine package framework dataclasses.

Dataclasses that are used to define a custom workflow and its steps.
"""
from dataclasses import dataclass, field
from typing import Optional

from .tasks import Task


@dataclass
class Step:
    step_id: str
    task_name: str
    target: Optional[str]
    start: Optional[bool] = None
    task_info: Optional[dict] = None
    description: Optional[str] = None
    groups: list[str] = field(default_factory=list)

    @property
    def task(self):
        return Task.tasks[self.task_name]


@dataclass
class Workflow:
    name: str
    steps: list[Step]

    def get_step(self, step_id):
        return next((step for step in self.steps if step.step_id == step_id), None)

    @property
    def first_step(self):
        return next(step for step in self.steps if step.start)



"""django_workflow_engine exceptions."""


class WorkflowImproperlyConfigured(Exception):
    pass


class WorkflowError(Exception):
    pass


class WorkflowNotAuthError(Exception):
    def __init__(self, step):
        self.step = step



import logging

from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied
from django.shortcuts import render, reverse, redirect
from django.urls import reverse_lazy
from django.http import JsonResponse, Http404
from django.views import View
from django.views.generic.edit import CreateView
from django.views.generic.edit import DeleteView
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView

from django_workflow_engine.forms import GovFormattedModelForm
from django_workflow_engine.tasks import TaskError
from django_workflow_engine.exceptions import WorkflowNotAuthError
from django_workflow_engine.executor import WorkflowExecutor
from django_workflow_engine.models import Flow, TaskRecord

logger = logging.getLogger(__name__)


class FlowListView(LoginRequiredMixin, ListView):
    model = Flow
    paginate_by = 100  # if pagination is desired
    ordering = "-started"


class FlowView(LoginRequiredMixin, DetailView):
    model = Flow


class FlowCreateForm(GovFormattedModelForm):
    class Meta:
        model = Flow
        fields = ["flow_name", "workflow_name"]


class FlowCreateView(LoginRequiredMixin, CreateView):
    model = Flow
    form_class = FlowCreateForm

    def get_success_url(self):
        return reverse("flow", args=[self.object.pk])

    def form_valid(self, form):
        obj = form.save(commit=False)
        obj.executed_by = self.request.user
        response = super().form_valid(form)
        executor = WorkflowExecutor(self.object)
        try:
            executor.run_flow(user=self.request.user)
        except WorkflowNotAuthError as e:
            logger.warning(f"{e}")
            raise PermissionDenied(f"{e}")
        return response


class FlowDeleteView(LoginRequiredMixin, DeleteView):
    model = Flow
    success_url = reverse_lazy('flow-list')


class FlowContinueView(LoginRequiredMixin, View):

    def __init__(self):
        super().__init__()
        self.flow = None
        self.step = None
        self.task = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.flow = Flow.objects.get(pk=kwargs.get("pk"))

        if self.flow.current_task_record:
            self.step = self.flow.workflow.get_step(
                self.flow.current_task_record.step_id
            )
            self.task = self.step.task(
                request.user, self.flow.current_task_record, self.flow
            )
        elif not self.flow.is_complete:
            logger.info(f"{self.flow.nice_name} is not complete")
            nodes = [step_to_node(self.flow, step) for step in self.flow.workflow.steps]
            for n in nodes:
                if n["data"]["done"] is None:
                    self.step = self.flow.workflow.get_step(n["data"]["id"])
                    self.task = self.step.task(
                        request.user, self.flow.current_task_record, self.flow
                    )
                    task_record, created = TaskRecord.objects.get_or_create(
                        flow=self.flow,
                        task_name=self.step.task_name,
                        step_id=self.step.step_id,
                        executed_by=None,
                        finished_at=None,
                        defaults={"task_info": self.step.task_info or {}},
                    )
                    break

    def get(self, request, pk, **kwargs):
        if not self.task:
            return redirect(reverse("flow-list"))

        context = self.get_context_data()

        template = self.task.template or "workflow/flow-continue.html"

        return render(request, template, context=context)

    def post(self, request, **kwargs):
        task_uuid = self.request.POST["uuid"]
        executor = WorkflowExecutor(self.flow)
        try:
            executor.run_flow(
                user=self.request.user, task_info=self.request.POST, task_uuid=task_uuid
            )
        except WorkflowNotAuthError as e:
            logger.warning(f"{e}")
            raise PermissionDenied(f"{e}")
        except TaskError as error:
            template = self.task.template or "workflow/flow-continue.html"

            context = self.get_context_data() | error.context

            return render(request, template, context=context)
        return redirect(reverse("flow", args=[self.flow.pk]))

    def get_context_data(self, **kwargs):
        context = {
            "flow": self.flow,
            "step": self.step,
            "task": self.task,
        }
        if not self.task.auto:
            context |= self.task.context()
        return context


class FlowDiagramView(LoginRequiredMixin, View):

    @staticmethod
    def get(request, pk, **kwargs):
        try:
            flow = Flow.objects.get(pk=pk)
        except Flow.DoesNotExist:
            raise Http404(f"Flow {pk} not found")
        elements = workflow_to_cytoscape_elements(flow)
        return JsonResponse({"elements": elements})


def workflow_to_cytoscape_elements(flow):
    nodes = [step_to_node(flow, step) for step in flow.workflow.steps]

    edges = []
    for step in flow.workflow.steps:
        targets = step.target if isinstance(step.target, list) else [step.target]
        for target in targets:
            if not target:
                continue

            edges.append(
                {
                    "data": {
                        "id": f"{step.step_id}{target}",
                        "source": step.step_id,
                        "target": target,
                    }
                }
            )

    return [*nodes, *edges]


def step_to_node(flow, step):
    latest_step_task = (
        flow.tasks.order_by("started_at").filter(step_id=step.step_id).last()
    )

    targets = []

    if step.target:
        targets = step.target if isinstance(step.target, list) else [step.target]

    end = not bool(targets)
    done = latest_step_task and bool(latest_step_task.finished_at)

    label = step.description or format_step_id(step.step_id)
    if end and done:
        label += " ✓"

    return {
        "data": {
            "id": step.step_id,
            "label": label,
            "start": step.start,
            "end": end,
            "decision": len(targets) > 1,
            "done": done,
            "current": latest_step_task and not latest_step_task.finished_at,
        }
    }


def format_step_id(step_id):
    # email_all_users -> Email all users
    return step_id.replace("_", " ").title()



from django import forms

from .task import Task


class EmailForm(forms.Form):
    subject = forms.CharField()
    message = forms.CharField(widget=forms.Textarea)
    from_email = forms.EmailField()
    recipient_list = forms.CharField()


class EmailFormTask(Task, input="email_form"):
    auto = False
    form = EmailForm

    def execute(self, task_info):
        form = self.form(task_info)

        if not form.is_valid():
            raise Exception(form.errors)

        form.cleaned_data["recipient_list"] = form.cleaned_data["recipient_list"].split(
            ","
        )

        return None, form.cleaned_data

    def context(self):
        return {"form": self.form(initial=self.task_record.task_info)}



from django import forms
from django.forms.widgets import (
    Textarea,
    Select,
    CheckboxInput,
    TextInput,
    EmailInput,
)


# TODO: maybe more to specific app
class GovFormattedModelForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.items():
            widget = field[1].widget
            if isinstance(widget, Textarea):
                widget.attrs.update({"class": "govuk-textarea"})
            elif isinstance(widget, Select):
                widget.attrs.update({"class": "govuk-select"})
            elif isinstance(widget, CheckboxInput):
                widget.attrs.update({"class": "govuk-checkboxes__input"})
            elif isinstance(widget, TextInput) or isinstance(widget, EmailInput):
                widget.attrs.update({"class": "govuk-input"})


class GovFormattedForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.items():
            widget = field[1].widget
            if isinstance(widget, Textarea):
                widget.attrs.update({"class": "govuk-textarea"})
            elif isinstance(widget, Select):
                widget.attrs.update({"class": "govuk-select"})
            elif isinstance(widget, CheckboxInput):
                widget.attrs.update({"class": "govuk-checkboxes__input"})
            elif isinstance(widget, TextInput) or isinstance(widget, EmailInput):
                widget.attrs.update({"class": "govuk-input"})



from django.core.mail import send_mail
from django.template import Context, Template

from .task import Task


class SendEmail(Task, input="send_email"):
    auto = True

    def execute(self, task_info):
        email_info = self.task_record.task_info | task_info

        message = Template(email_info["message"])
        context = Context(
            self.flow.flow_info
            | email_info
            | {"flow": self.flow, "task": self.task_record}
        )

        successfully_sent = send_mail(
            subject=email_info["subject"],
            message=message.render(context),
            from_email=email_info["from_email"],
            recipient_list=email_info["recipient_list"],
            fail_silently=False,
        )

        return None, {"successfully_sent": successfully_sent}



"""django_workflow_engine package urls."""

from django.urls import path

import django_workflow_engine.views as workflow_views


urlpatterns = [
    path("", workflow_views.FlowListView.as_view(), name="flow-list"),
    path("new", workflow_views.FlowCreateView.as_view(), name="flow-create"),
    path("delete/<int:pk>/", workflow_views.FlowDeleteView.as_view(), name="flow-delete"),
    path("<int:pk>/", workflow_views.FlowView.as_view(), name="flow"),
    path(
        "<int:pk>/continue",
        workflow_views.FlowContinueView.as_view(),
        name="flow-continue",
    ),
    path(
        "<int:pk>/diagram",
        workflow_views.FlowDiagramView.as_view(),
        name="flow-diagram",
    ),
]



"""Django workflow Task.

A task is an instance of a django_workflow_engine.dataclasses.Step
"""


class TaskError(Exception):
    def __init__(self, message, context):
        self.message = message
        self.context = context


class Task:
    tasks = {}
    template = None

    def __init__(self, user, task_record, flow):
        self.user = user
        self.task_record = task_record
        self.flow = flow

    def setup(self, task_info):
        pass

    def execute(self, task_info):
        raise NotImplementedError

    def log(self, message):
        self.task_record.log.create(message=message)

    def __init_subclass__(cls, /, input, **kwargs):  # noqa
        super().__init_subclass__(**kwargs)
        cls.tasks[input] = cls



from distutils.core import setup

import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="django-workflow-engine",
    version="0.0.2",
    packages=setuptools.find_packages(),
    author="DIT Live Service Team",
    author_email="live.services@digital.trade.gov.uk",
    url="https://github.com/uktrade/django-workflow-engine",
    description="Lightweight, reusable workflow engine for Django applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "django>=2.2.24",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from django.utils import timezone

from .exceptions import WorkflowError, WorkflowNotAuthError
from django_workflow_engine.models import TaskRecord


class WorkflowExecutor:
    def __init__(self, flow):
        self.flow = flow

    def run_flow(self, user, task_info=None, task_uuid=None):
        """Run the workflow.

        We identify the current step and execute workflow steps until a manual
        step is encountered (e.g. complete a form) or the workflow is exhausted.

        :param (User) user: User requesting to run a flow step.
        :param (dict) task_info: Additional task info.
        :param (str) task_uuid: Task record UUID.
        :returns (TaskRecord): New or updated task record.
        """
        if task_info is None:
            task_info = {}

        # TODO: might be a race condition
        if self.flow.started and not task_uuid:
            raise WorkflowError("Flow already started")

        current_step = self.get_current_step(self.flow, task_uuid)
        while current_step:
            task_record, created = TaskRecord.objects.get_or_create(
                flow=self.flow,
                task_name=current_step.task_name,
                step_id=current_step.step_id,
                executed_by=None,
                finished_at=None,
                defaults={"task_info": current_step.task_info or {}},
            )

            task = current_step.task(user, task_record, self.flow)

            task.setup(task_info)

            # the next task has a manual step
            if not task.auto and created:
                return task_record

            self.check_authorised(user, current_step)  # Raises if user not authorised for step
            target, task_output = task.execute(task_info)

            # TODO: check target against step target

            task_record.finished_at = timezone.now()
            task_record.save()
            self.flow.save()

            current_step = next(
                (
                    step
                    for step in self.flow.workflow.steps
                    if step.step_id == (target or current_step.target)
                ),
                None,
            )

            task_info = task_output

        self.flow.finished = timezone.now()
        self.flow.save()

        return task_record

    @staticmethod
    def get_current_step(flow, task_uuid=None):
        """Get the current step.

        If a task uuid is provided we retrieve the related task record/step
        otherwise use the workflow's designated first step.

        :param (Flow) flow: the workflow.
        :param (str) task_uuid: UUID of the current task or None.
        :returns (Step): A workflow step.
        """
        if task_uuid:
            current_step = flow.workflow.get_step(
                TaskRecord.objects.get(uuid=task_uuid).step_id
            )
        else:
            current_step = flow.workflow.first_step
            flow.started = timezone.now()
            flow.save()
        return current_step

    @staticmethod
    def check_authorised(user, step):
        """Check user is authorised.

        Check that a user is authorised to execute a workflow step.

        :param (User) user: Workflow user.
        :param (Step) step: Flow step.
        :raises (WorkflowNotAuthError): If user is not authorised.
        """
        user_groups = set(g.name for g in user.groups.all())
        required_groups = set(step.groups)
        if user_groups.intersection(required_groups):
            return
        msg = (
            f"User '{user}' is not authorised to execute the "
            f"step: {step.task_name}"
        )
        raise WorkflowNotAuthError(msg)
